@@ -2,7 +2,7 @@ import hmac
 import importlib
 import inspect
 import logging
-from typing import Any, Callable, Dict, Optional, Type, Union
+from typing import Any, Callable, Dict, List, Optional, Type, Union
 
 import orjson
 import shortuuid
@@ -114,12 +114,13 @@ class UnicornTemplateResponse(TemplateResponse):
 
         if self.init_js:
             script_tag = soup.new_tag("script")
-            init_data = {
+            init = {
                 "id": self.component_id,
                 "name": self.component_name,
+                "data": orjson.loads(self.frontend_context_variables),
             }
-            init = orjson.dumps(init_data).decode("utf-8")
-            script_tag.string = f"if (typeof Unicorn === 'undefined') {{ console.error('Unicorn is missing. Do you need {{% load unicorn %}} or {{% unicorn-scripts %}}?') }} else {{ Unicorn.setData({self.frontend_context_variables}); Unicorn.componentInit({init}); }}"
+            init = orjson.dumps(init).decode("utf-8")
+            script_tag.string = f"if (typeof Unicorn === 'undefined') {{ console.error('Unicorn is missing. Do you need {{% load unicorn %}} or {{% unicorn-scripts %}}?') }} else {{ Unicorn.componentInit({init}); }}"
             root_element.insert_after(script_tag)
 
         rendered_template = UnicornTemplateResponse._desoupify(soup)
@@ -158,6 +159,8 @@ class UnicornView(TemplateView):
     # Caches to reduce the amount of time introspecting the class
     _methods_cache = None
     _attribute_name_cache = None
+    _form_cache = None
+    _hook_methods_cache: List[str] = []
 
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
@@ -170,6 +173,7 @@ class UnicornView(TemplateView):
         if "request" in kwargs:
             self.setup(kwargs["request"])
 
+        self.errors = {}
         self._set_default_template_name()
         self._set_caches()
 
@@ -195,13 +199,43 @@ class UnicornView(TemplateView):
         Setup some initial "caches" to prevent Python from having to introspect
         a component UnicornView for methods and properties multiple times.
         """
-        self._methods_cache = self._methods()
         self._attribute_names_cache = self._attribute_names()
+        self._set_hook_methods_cache()
+        self._methods_cache = self._methods()
 
     def mount(self):
         """
-        A hook that will always get called when a component is created.
-        By default a no-op, components can override it if needed.
+        Hook that gets called when a component is first created.
+        """
+        pass
+
+    def hydrate(self):
+        """
+        Hook that gets called when a component's data is hydrated.
+        """
+        pass
+
+    def updating(self, name, value):
+        """
+        Hook that gets called when a component's data is about to get updated.
+        """
+        pass
+
+    def updated(self, name, value):
+        """
+        Hook that gets called when a component's data is updated.
+        """
+        pass
+
+    def calling(self, name, args):
+        """
+        Hook that gets called when a component's method is about to get called.
+        """
+        pass
+
+    def called(self, name, args):
+        """
+        Hook that gets called when a component's method is called.
         """
         pass
 
@@ -209,6 +243,9 @@ class UnicornView(TemplateView):
         """
         Renders a UnicornView component with the public properties available. Delegates to a
         UnicornTemplateResponse to actually render a response.
+
+        Args:
+            param init_js: Whether or not to include the Javascript required to initialize the component.
         """
         frontend_context_variables = self.get_frontend_context_variables()
 
@@ -273,24 +310,98 @@ class UnicornView(TemplateView):
             raise TypeError
 
         frontend_context_variables = {}
-        frontend_context_variables.update(self._attributes())
+        attributes = self._attributes()
+        frontend_context_variables.update(attributes)
+
+        # Add cleaned values to `frontend_content_variables` based on the widget in form's fields
+        form = self._get_form(attributes)
+
+        if form:
+            for key in attributes.keys():
+                if key in form.fields:
+                    field = form.fields[key]
+
+                    if key in form.cleaned_data:
+                        cleaned_value = form.cleaned_data[key]
+                        value = field.widget.format_value(cleaned_value)
+                        frontend_context_variables[key] = value
+
         encoded_frontend_context_variables = orjson.dumps(
-            frontend_context_variables, default=_json_serializer
+            frontend_context_variables, default=_json_serializer,
         ).decode("utf-8")
 
         return encoded_frontend_context_variables
 
+    def _get_form(self, data, use_cache=True):
+        if self._form_cache and use_cache:
+            return self._form_cache
+
+        if hasattr(self, "form_class"):
+            try:
+                form = self.form_class(data)
+                form.is_valid()
+
+                self._form_cache = form
+                return form
+            except Exception as e:
+                logger.exception(e)
+
     def get_context_data(self, **kwargs):
         """
         Overrides the standard `get_context_data` to add in publicly available
         properties and methods.
         """
+
         context = super().get_context_data(**kwargs)
-        context.update(self._attributes())
+
+        attributes = self._attributes()
+        context.update(attributes)
         context.update(self._methods())
+        context.update({"unicorn": {"errors": self.errors}})
 
         return context
 
+    def validate(self, model_names: List = None) -> Dict:
+        """
+        Validates the data using the `form_class` set on the component.
+
+        Args:
+            model_names: Only include validation errors for specified fields. If none, validate everything.
+        """
+        # TODO: Handle form.non_field_errors()?
+
+        data = self._attributes()
+        form = self._get_form(data, use_cache=False)
+
+        if form:
+            form_errors = form.errors.get_json_data(escape_html=True)
+
+            if model_names is not None:
+                # This code is confusing, but handles this use-case:
+                # the component has two models, one that starts with an error and one
+                # that is valid. Validating the valid one should not show an error for
+                # the invalid one. Only after the invalid field is updated, should the
+                # error show up and persist, even after updating the valid form.
+                if self.errors:
+                    keys_to_remove = []
+
+                    for key, value in self.errors.items():
+                        if key in form_errors:
+                            self.errors[key] = value
+                        else:
+                            keys_to_remove.append(key)
+
+                    for key in keys_to_remove:
+                        self.errors.pop(key)
+
+                for key, value in form_errors.items():
+                    if key in model_names:
+                        self.errors[key] = value
+            else:
+                self.errors.update(form_errors)
+
+        return self.errors
+
     def _attribute_names(self):
         """
         Gets publicly available attribute names. Cached in `_attribute_names_cache`.
@@ -316,7 +427,29 @@ class UnicornView(TemplateView):
         return attributes
 
     def _set_property(self, name, value):
-        setattr(self, name, value)
+        # Get the correct value type by using the form if it is available
+        data = self._attributes()
+        data[name] = value
+        form = self._get_form(data, use_cache=False)
+
+        if form and name in form.fields and name in form.cleaned_data:
+            value = form.cleaned_data[name]
+
+        updating_function_name = f"updating_{name}"
+        if hasattr(self, updating_function_name):
+            getattr(self, updating_function_name)(value)
+
+        try:
+            setattr(self, name, value)
+
+            updated_function_name = f"updated_{name}"
+
+            if hasattr(self, updated_function_name):
+                getattr(self, updated_function_name)(value)
+        except AttributeError:
+            logger.error(
+                f"'{name}' attribute on '{self.component_name}' component could not be set. Is it a @property without a setter?"
+            )
 
     def _methods(self) -> Dict[str, Callable]:
         """
@@ -336,6 +469,18 @@ class UnicornView(TemplateView):
 
         return methods
 
+    def _set_hook_methods_cache(self) -> None:
+        self._hook_methods_cache = []
+
+        for attribute_name in self._attribute_names_cache:
+            updating_function_name = f"updating_{attribute_name}"
+            updated_function_name = f"updated_{attribute_name}"
+            hook_function_names = [updating_function_name, updated_function_name]
+
+            for function_name in hook_function_names:
+                if hasattr(self, function_name):
+                    self._hook_methods_cache.append(function_name)
+
     def _is_public(self, name: str) -> bool:
         """
         Determines if the name should be sent in the context.
@@ -354,9 +499,6 @@ class UnicornView(TemplateView):
             "template_name",
             "dispatch",
             "id",
-            "component_id",
-            "component_name",
-            "get_frontend_context_variables",
             "get",
             "get_context_data",
             "get_template_names",
@@ -364,16 +506,32 @@ class UnicornView(TemplateView):
             "http_method_not_allowed",
             "options",
             "setup",
-            "render",
             "fill",
+            # Component methods
+            "component_id",
+            "component_name",
             "mount",
+            "hydrate",
+            "updating",
+            "update",
+            "calling",
+            "called",
+            "validate",
+            "get_frontend_context_variables",
+            "errors",
+            "updated",
         )
         excludes = []
 
         if hasattr(self, "Meta") and hasattr(self.Meta, "exclude"):
             excludes = self.Meta.exclude
 
-        return not (name.startswith("_") or name in protected_names or name in excludes)
+        return not (
+            name.startswith("_")
+            or name in protected_names
+            or name in self._hook_methods_cache
+            or name in excludes
+        )
 
     @staticmethod
     def create(
@@ -402,31 +560,26 @@ class UnicornView(TemplateView):
             component = views_cache[component_name](
                 component_name=component_name, component_id=component_id
             )
-            component.mount()
 
             if not use_cache:
                 #  Re-initializes custom classes so that `reset` magic method will "clear" them as expected
-
-                NON_CONSTRUCTABLE_TYPES = (
-                    str,
-                    int,
-                    dict,
-                    list,
-                    Model,
-                    QuerySet,
-                )
                 _attributes = component._attributes()
 
                 for attribute_name in _attributes:
                     attribute = _attributes[attribute_name]
 
-                    # TODO: Not sure if there is a better way to do this. Could just check for UnicornField?
-                    if not any(
-                        filter(
-                            lambda t: isinstance(attribute, t), NON_CONSTRUCTABLE_TYPES
-                        )
-                    ):
-                        component._set_property(attribute_name, attribute.__class__())
+                    if isinstance(attribute, UnicornField):
+                        try:
+                            component._set_property(
+                                attribute_name, attribute.__class__()
+                            )
+                        except TypeError:
+                            logger.warn(
+                                f"Resetting '{attribute_name}' attribute failed because it could not be constructed."
+                            )
+                            pass
+
+                component.hydrate()
 
             if component_id:
                 key = f"{component_name}-{component_id}"
@@ -476,6 +629,7 @@ class UnicornView(TemplateView):
                     component_name=component_name, id=component_id
                 )
                 component.mount()
+                component.hydrate()
 
                 views_cache[component_name] = component_class
 
